package com.sogou.shortvideo

import android.bluetooth.BluetoothAdapter
import android.bluetooth.BluetoothHeadset
import android.bluetooth.BluetoothProfile
import android.content.BroadcastReceiver
import android.content.Context
import android.content.Intent
import android.content.IntentFilter
import android.net.Uri
import android.os.Build
import android.util.Log
import android.util.LongSparseArray
import android.view.Surface
import com.google.android.exoplayer2.*
import com.google.android.exoplayer2.Player.*
import com.google.android.exoplayer2.audio.AudioAttributes
import com.google.android.exoplayer2.extractor.DefaultExtractorsFactory
import com.google.android.exoplayer2.source.ExtractorMediaSource
import com.google.android.exoplayer2.source.MediaSource
import com.google.android.exoplayer2.source.dash.DashMediaSource
import com.google.android.exoplayer2.source.dash.DefaultDashChunkSource
import com.google.android.exoplayer2.source.hls.HlsMediaSource
import com.google.android.exoplayer2.source.smoothstreaming.DefaultSsChunkSource
import com.google.android.exoplayer2.source.smoothstreaming.SsMediaSource
import com.google.android.exoplayer2.trackselection.DefaultTrackSelector
import com.google.android.exoplayer2.upstream.*
import com.google.android.exoplayer2.upstream.cache.CacheDataSource.FLAG_IGNORE_CACHE_ON_ERROR
import com.google.android.exoplayer2.upstream.cache.CacheDataSourceFactory
import com.google.android.exoplayer2.util.Util
import io.flutter.plugin.common.EventChannel
import io.flutter.plugin.common.MethodCall
import io.flutter.plugin.common.MethodChannel.Result
import io.flutter.plugin.common.PluginRegistry
import io.flutter.view.TextureRegistry

class SgExoPlayerPlugin : IVideoPlugin{

    private val videoPlayers: LongSparseArray<VideoPlayer> = LongSparseArray()
    private var receiver: BroadcastReceiver? = null
    private var registrar: PluginRegistry.Registrar? = null

    internal class VideoPlayer(
            context: Context,
            private val eventChannel: EventChannel,
            private val textureEntry: TextureRegistry.SurfaceTextureEntry) {

        private val TAG: String = "ExoVideoPlayer"

        private val exoPlayer: SimpleExoPlayer?

        private var surface: Surface? = null

        private var id: Long? = null

        private val eventSink = QueuingEventSink()

        private var isInitialized = false

        // 弱网或seek造成的缓冲
        private var isBuffering = false

        val position: Long
            get() = exoPlayer!!.currentPosition

        init {
            val trackSelector = DefaultTrackSelector()
            val loadControl = DefaultLoadControl.Builder()
                    .setAllocator(DefaultAllocator(true, C.DEFAULT_BUFFER_SEGMENT_SIZE))
                    .setBufferDurationsMs(1200, 3000, 500, 1000)
                    .createDefaultLoadControl()
            exoPlayer = ExoPlayerFactory.newSimpleInstance(context, trackSelector, loadControl)

            eventChannel.setStreamHandler(
                    object : EventChannel.StreamHandler {
                        override fun onListen(o: Any?, sink: EventChannel.EventSink?) {
                            eventSink.setDelegate(sink)
                        }

                        override fun onCancel(o: Any?) {
                            eventSink.setDelegate(null)
                        }
                    })


            id = textureEntry.id()
            surface = Surface(textureEntry.surfaceTexture())
            exoPlayer?.setVideoSurface(surface)
            setAudioAttributes(exoPlayer)
        }

        private fun buildMediaSource(
                uri: Uri, mediaDataSourceFactory: DataSource.Factory, formatHint: String?, context: Context): MediaSource {
            val type: Int
            if (formatHint == null) {
                type = Util.inferContentType(uri.lastPathSegment)
            } else {
                when (formatHint) {
                    FORMAT_SS -> type = C.TYPE_SS
                    FORMAT_DASH -> type = C.TYPE_DASH
                    FORMAT_HLS -> type = C.TYPE_HLS
                    FORMAT_OTHER -> type = C.TYPE_OTHER
                    else -> type = -1
                }
            }
            when (type) {
                C.TYPE_SS -> return SsMediaSource.Factory(
                        DefaultSsChunkSource.Factory(mediaDataSourceFactory),
                        DefaultDataSourceFactory(context, null, mediaDataSourceFactory))
                        .createMediaSource(uri)
                C.TYPE_DASH -> return DashMediaSource.Factory(
                        DefaultDashChunkSource.Factory(mediaDataSourceFactory),
                        DefaultDataSourceFactory(context, null, mediaDataSourceFactory))
                        .createMediaSource(uri)
                C.TYPE_HLS -> return HlsMediaSource.Factory(mediaDataSourceFactory).createMediaSource(uri)
                C.TYPE_OTHER -> return ExtractorMediaSource.Factory(mediaDataSourceFactory)
                        .setExtractorsFactory(DefaultExtractorsFactory())
                        .createMediaSource(uri)
                else -> {
                    throw IllegalStateException("Unsupported type: $type")
                }
            }
        }

        fun sendBufferingUpdate() {
            Log.e(TAG, "$id -> pending true")
            val event = HashMap<String, Any>()
            event["event"] = "pending"
            event["isPending"] = true
            eventSink.success(event)
        }

        fun sendBufferingReady() {
            Log.e(TAG, "$id -> pending false")
            val event = HashMap<String, Any>()
            event["event"] = "pending"
            event["isPending"] = false
            eventSink.success(event)
        }

        fun play() {
            exoPlayer?.playWhenReady = true
            sendPlayed()
        }

        fun pause() {
            exoPlayer?.playWhenReady = false
            sendPaused()
        }

        fun setLooping(value: Boolean) {
            exoPlayer?.repeatMode = if (value) REPEAT_MODE_ALL else REPEAT_MODE_OFF
        }

        fun setVolume(value: Double) {
            val bracketedValue = Math.max(0.0, Math.min(1.0, value)).toFloat()
            exoPlayer?.volume = bracketedValue
        }

        fun seekTo(location: Long) {
            exoPlayer?.seekTo(location)
        }

        fun stop() {
            exoPlayer?.stop(true)
            sendStoped()
        }

        private fun sendPlayed() {
            Log.e(TAG, "$id -> sendPlayed")
            val event = HashMap<String, Any>()
            event["event"] = "playing"
            eventSink.success(event)
        }

        private fun sendPaused() {
            Log.e(TAG, "$id -> sendPaused")
            val event = HashMap<String, Any>()
            event["event"] = "paused"
            eventSink.success(event)
        }

        private fun sendStoped() {
            Log.e(TAG, "$id -> sendStoped")
            val event = HashMap<String, Any>()
            event["event"] = "created"
            eventSink.success(event)
        }

        private fun sendInitialized() {
            if (isInitialized) {
                val event = HashMap<String, Any>()
                event["event"] = "initialized"
                event["duration"] = exoPlayer!!.duration

                exoPlayer?.videoFormat?.also {
                    var width = it.width
                    var height = it.height
                    val rotationDegrees = it.rotationDegrees
                    // Switch the width/height if video was taken in portrait mode
                    if (rotationDegrees == 90 || rotationDegrees == 270) {
                        width = it.height
                        height = it.width
                    }
                    event["width"] = width
                    event["height"] = height
                }
                eventSink.success(event)
            }
        }

        var previousListener : EventListener? = null

        fun resetToNewDatasource(context: Context, dataSource: String,
                autoPlay :Boolean, id: Long) {

            var start = System.currentTimeMillis()

            exoPlayer?.stop(true)
            val uri = Uri.parse(dataSource)

            val dataSourceFactory: DataSource.Factory
            if (isHTTP(uri)) {
                dataSourceFactory = DefaultHttpDataSourceFactory(
                        "ExoPlayer", null,
                        DefaultHttpDataSource.DEFAULT_CONNECT_TIMEOUT_MILLIS,
                        DefaultHttpDataSource.DEFAULT_READ_TIMEOUT_MILLIS,
                        true)
            } else {
                dataSourceFactory = DefaultDataSourceFactory(context, "ExoPlayer")
            }

            val cacheDataSourceFactory = CacheDataSourceFactory(ExoVideoCache.get(), dataSourceFactory, FLAG_IGNORE_CACHE_ON_ERROR)

            val mediaSource = buildMediaSource(uri, cacheDataSourceFactory, "other", context)
            exoPlayer?.prepare(mediaSource)

            if (previousListener != null){
                //复用player时清除之前的listener和一些标记位
                isInitialized = false
                isBuffering = false
                exoPlayer?.removeListener(previousListener)
            }

            var eventListener = object : EventListener {

                override fun onPlayerStateChanged(playWhenReady: Boolean, playbackState: Int) {
                    if (playbackState == STATE_BUFFERING) {
                        isBuffering = true
                        sendBufferingUpdate()
                    } else if (playbackState == STATE_READY) {
                        if (isInitialized){
                            if (isBuffering){
                                isBuffering = false
                                sendBufferingReady()
                            }
                        }else{
                            isInitialized = true
                            sendInitialized()
                            if (autoPlay){
                                play()
                            }
                        }
                    } else if (playbackState == STATE_ENDED) {
                        //REPEAT_MODE_OFF回调该状态
                        //REPEAT_MODE_ONE、REPEAT_MODE_ALL不会回调该状态
                        val event = HashMap<String, Any>()
                        event["event"] = "completed"
                        eventSink.success(event)
                    }
                }

                override fun onPlayerError(error: ExoPlaybackException) {
                    if (eventSink != null) {
                        eventSink.error("VideoError", "Video player had error $error", "")
                    }
                }
            }

            exoPlayer?.addListener(eventListener)
            previousListener = eventListener

            exoPlayer?.setVideoListener(object : SimpleExoPlayer.VideoListener {
                override fun onRenderedFirstFrame() {
                    Log.i(TAG, "$id -> $exoPlayer first frame rendered")
                }
            })
        }

        fun dispose() {
            if (isInitialized) {
                exoPlayer?.stop()
            }
            textureEntry.release()
            eventChannel.setStreamHandler(null)
            if (surface != null) {
                surface?.release()
            }
            if (exoPlayer != null) {
                exoPlayer?.release()
            }
        }

        fun debug(debugCode: Int) {
            if (debugCode == 1) {
                val event = java.util.HashMap<String, Any>()
                event["event"] = "occurError"
                event["error"] = "Video player had error manullay"
                eventSink.success(event)
            }
        }

        companion object {
            private const val FORMAT_SS = "ss"
            private const val FORMAT_DASH = "dash"
            private const val FORMAT_HLS = "hls"
            private const val FORMAT_OTHER = "other"

            private fun isHTTP(uri: Uri?): Boolean {
                if (uri == null || uri.scheme == null) {
                    return false
                }
                val scheme = uri.scheme
                return scheme == "http" || scheme == "https"
            }

            private fun setAudioAttributes(exoPlayer: SimpleExoPlayer) {
                if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP) {
                    exoPlayer.audioAttributes = AudioAttributes.Builder().setContentType(C.CONTENT_TYPE_MOVIE).build()
                } else {
                    exoPlayer.audioStreamType = C.STREAM_TYPE_MUSIC
                }
            }
        }

    }

    private fun disposeAllPlayers() {
        for (i in 0 until videoPlayers.size()) {
            videoPlayers.valueAt(i).dispose()
        }
        videoPlayers.clear()
    }

    override fun onCreate(registrar: PluginRegistry.Registrar) {
        this.registrar = registrar
        receiver = createReceiver()
        var intentFilter = IntentFilter()
        intentFilter.addAction(Intent.ACTION_HEADSET_PLUG)
        intentFilter.addAction(BluetoothHeadset.ACTION_CONNECTION_STATE_CHANGED)
        registrar.context()
                .registerReceiver(receiver, intentFilter)
    }

    private fun createReceiver(): BroadcastReceiver {
        return object : BroadcastReceiver() {

            var headsetIn = false
            var bluetoothIn = false

            override fun onReceive(context: Context, intent: Intent) {

                var action = intent.action

                if ("android.intent.action.HEADSET_PLUG" == action) {
                    if (intent.hasExtra("state")){
                        if (intent.getIntExtra("state", 0) == 0){
                            if (headsetIn){
                                headsetIn = false
                                for (i in 0 until videoPlayers.size()) {
                                    videoPlayers.valueAt(i).pause()
                                }
                            }
                        } else if (intent.getIntExtra("state", 0) == 1){
                            headsetIn = true
                        }
                    }
                } else if (BluetoothHeadset.ACTION_CONNECTION_STATE_CHANGED == action) {
                    var adapter = BluetoothAdapter.getDefaultAdapter()
                    if (BluetoothProfile.STATE_CONNECTED == adapter.getProfileConnectionState(BluetoothProfile.HEADSET)) {
                        bluetoothIn = true
                    }else if (BluetoothProfile.STATE_DISCONNECTED == adapter.getProfileConnectionState(BluetoothProfile.HEADSET)){
                        if (bluetoothIn){
                            bluetoothIn = false
                            for (i in 0 until videoPlayers.size()) {
                                videoPlayers.valueAt(i).pause()
                            }
                        }
                    }
                }
            }
        }
    }

    override fun onDestroy() {
        disposeAllPlayers()
        Thread{
            ExoVideoCache.deleteCacheFile()
        }.start()
        registrar?.context()?.unregisterReceiver(receiver)
        receiver = null
    }

    override fun dispatchCall(registrar: PluginRegistry.Registrar, call: MethodCall, result: Result) {
        val textures = registrar.textures()
        if (textures == null) {
            result.error("no_activity", "video_player plugin requires a foreground activity", null)
            return
        }
        when (call.method) {
            "init" -> disposeAllPlayers()
            "create" -> {
                val handle = textures.createSurfaceTexture()
                val eventChannel = EventChannel(
                        registrar.messenger(), "sogou.flutter/videoPlayer/videoEvents" + handle.id())

                Log.i("SgVideoPlayerPlugin", " Create NEW " + handle.id() + "-->" + call.argument("uri"))

                val player: VideoPlayer

                player = VideoPlayer(
                        registrar.context(), eventChannel, handle)
                val reply = HashMap<String, Any>()
                reply["textureId"] = handle.id()
                result.success(reply)
                videoPlayers.put(handle.id(), player)
                Log.e("SgVideoPlayerPlugin", "" + handle.id() + "-->" + call.method)
            }
            else -> {
                val textureId = (call.argument<Any>("textureId") as Number).toLong()
                val player = videoPlayers.get(textureId)
                if (player == null) {
                    result.error(
                            "Unknown textureId",
                            "No video player associated with texture id $textureId", null)
                    return
                }
                onMethodCall(registrar.context(), call, result, textureId, player)
            }
        }

    }

    private fun onMethodCall(context: Context, call: MethodCall, result: Result, textureId: Long, player: VideoPlayer?) {
        Log.i("SgVideoPlayerPlugin", "" + textureId + "-->" + call.method)
        when (call.method) {
            "setLooping" -> {
                player?.setLooping(call.argument<Boolean>("looping")!!)
                result.success(null)
            }
            "setVolume" -> {
                player?.setVolume(call.argument<Double>("volume")!!)
                result.success(null)
            }
            "play" -> {
                player?.play()
                result.success(null)
            }
            "pause" -> {
                player?.pause()
                result.success(null)
            }
            "seekTo" -> {
                val location = (call.argument<Any>("location") as Number).toLong()
                player?.seekTo(location)
                result.success(null)
            }
            "position" -> {
                result.success(player?.position)
                player?.sendBufferingUpdate()
            }
            "dispose" -> {
                player?.dispose()
                videoPlayers.remove(textureId)
                result.success(null)
            }
            "reset" -> {
                // 如果有错误，记得清理错误标记
                val newDataSource = (call.argument<Any>("uri") as String)
                val autoPlay = (call.argument<Any>("autoPlay") as Boolean)

                player?.resetToNewDatasource(
                        context, newDataSource, autoPlay, textureId)
                result.success(null)
            }
            "stop" -> {
                player?.stop()
                result.success(null)
            }
            "debug" -> {
                val debugCode = (call.argument<Any>("debugCode") as Number).toInt()
                player?.debug(debugCode)
                result.success(null)
            }
            else -> result.notImplemented()
        }
    }
}