import 'dart:async';
import 'dart:ui';

import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';
import 'package:scoped_model/scoped_model.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:short_video/base/i_log.dart';
import 'package:short_video/base/i_net_http.dart';
import 'package:short_video/base/i_store_sp.dart';
import 'package:short_video/build_config.dart';
import 'package:short_video/locale/localization.dart';
import 'package:short_video/main_screen.dart';
import 'package:short_video/model_main.dart';
import 'package:short_video/network/root_url.dart';
import 'package:short_video/network/sg_http_client.dart';
import 'package:short_video/platform/sg_operation.dart';
import 'package:short_video/sg_log.dart';
import 'package:short_video/storage/imp_base_store_sp.dart';
import 'package:short_video/test_screen.dart';
import 'package:short_video/util/constant.dart';
import 'package:short_video/util/crash_handler.dart';

import 'network/i_req_base.dart';

Future<Null> main() async {

  // This captures errors reported by the Flutter framework.
  FlutterError.onError = (FlutterErrorDetails details) async {
    if (BuildConfig().isInDebugMode()) {
      // In development mode simply print to console.
      FlutterError.dumpErrorToConsole(details);
    } else {
      // In production mode report to the application zone to record
      Zone.current.handleUncaughtError(details.exception, details.stack);
    }
  };

  runZoned<Future<Null>>(() async {
    initApp();
  }, onError: (error, stackTrace) async{
    CrashHandler.captureException(exception: error, stackTrace: stackTrace);
  });
}

void initApp() async{
  SgLog.initLog();

  if (BuildConfig().isInDebugMode()) {
    ILog.lg.i("App running in DEBUG mode ");
  } else {
    ILog.lg.i("App running in RELEASE mode ");
  }

  // 事先调用此函数，否则初始化channel会出错；
  WidgetsFlutterBinding.ensureInitialized();

  // 初始化顺序不得乱： key-value本地存储 -> 初始化设备信息 -> 初始化网络组件（包含加载本地url）
  await initSharePref();
  // 如果需要，此处可以初始化数据操作相关： [IStoreDb]
  await initDeviceInfo();
  await initNetwork();

  ///与native约定以URI格式传参
  ///目前的参数有video json字符串、点击的频道、小视频播放页翻页形式
  ///flutter://navigation?video=xxx&channel=小视频&orderType=1
  String schema = window.defaultRouteName;
  Uri uri;
  String videoJson;
  int orderType = MainModel.ORDER_TYPE_MANUAL;
  String channel = "推荐";

  try{
    uri = Uri.parse(schema);
    videoJson = uri.queryParameters['video'];
    orderType = int.parse(uri.queryParameters['orderType']);
    channel = uri.queryParameters['channel'];
  } on CastError catch (stack){

  } on Error catch(stack){

  }

  Constant.CHANNEL = channel;

  var app = MaterialApp(
    debugShowCheckedModeBanner: false,
    theme: ThemeData(
        primaryColor: Color(0xff00ff00), accentColor: Color(0xff96b6f4)),
    localizationsDelegates: [
      LoanLocalizationsDelegate(),
    ],
    home: MainScreen(),
  );

  // 创建scopedmodel
  var model = MainModel(videoJson, orderType);
  var scopedMode = ScopedModel<MainModel>(
    model: model,
    child: app,
  );


  runApp(scopedMode);
//  app = MaterialApp(
//    debugShowCheckedModeBanner: false,
//    theme: ThemeData(
//        primaryColor: Color(0xff00ff00), accentColor: Color(0xff96b6f4)),
//    localizationsDelegates: [
//      LoanLocalizationsDelegate(),
//    ],
//    home: TestScreen(),
//  );
//
//  runApp(app);
}

Future<bool> initSharePref() async {
  var spInstance = await SharedPreferences.getInstance();
  IStoreSp.resetSp(ImpBaseStoreSp(spInstance));
  return true;
}

Future<bool> initNetwork() async {
  // 初始化网络http请求
  INetHttp.resetHttp(SogouClient());
  // 加载 host
  return UrlRoot.onAppLaunch();
}

/// 初始化设备信息，如果需要的话
Future<bool> initDeviceInfo() async {
  ISgReqBusiness.paraCommon = await SgOperation.getSysInfo();
  return true;
}
