import 'dart:async';
import 'dart:io';

import 'package:connectivity/connectivity.dart';
import 'package:flutter/material.dart';
import 'package:flutter/rendering.dart';
import 'package:scoped_model/scoped_model.dart';
import 'package:short_video/base/i_log.dart';
import 'package:short_video/model_main.dart';
import 'package:short_video/util/toast.dart';
import 'package:short_video/widget/common/sg_page_view.dart';
import 'package:short_video/widget/error_widget.dart';
import 'package:short_video/widget/page_item_screen.dart';
import 'package:short_video/widget/user_guid_widget.dart';

import 'model_item.dart';
import 'platform/sg_operation.dart';
import 'util/constant.dart';

class MainScreen extends StatefulWidget {
  static final GlobalKey<_MainScreenState> mainScreenKey =
      GlobalKey<_MainScreenState>();

  MainScreen() : super(key: mainScreenKey);

  @override
  State<StatefulWidget> createState() {
    return _MainScreenState();
  }
}

class _MainScreenState extends State<MainScreen> with WidgetsBindingObserver {
  StreamSubscription subscriptionNetwork;
  SgPageController _pageController;
  Timer _timerUserGuid;
  // 网络状态变更，只显示一次
  bool _networkToastShown;
  // 当前页面是否被隐藏
  bool isHiden;
  @override
  void initState() {
    print("TIME- _MainScreenState::initState");
    super.initState();
    isHiden = false;

    // 设置图片缓冲区 5M
    imageCache.maximumSizeBytes = 5 * 1024 * 1024 ;

    _networkToastShown = false;
    // 初次进入，判断下网络状态
    Connectivity().checkConnectivity().then(_onNetworkStates);
    // 初始化网络状态监听
    subscriptionNetwork =
        Connectivity().onConnectivityChanged.listen(_onNetworkStates);
    _pageController = SgPageController();

    var model = MainModel.of(context);
    model.resetPageController(_pageController);
    if (model.needUserGuid()) {
      // 定时清理 user guid
      _timerUserGuid = Timer(Duration(seconds: 10), () {
        model.resetUserGuid(false);
        _timerUserGuid.cancel();
        _timerUserGuid = null;
      });
    }

    // 监听app被切换到后台等信息
    WidgetsBinding.instance.addObserver(this);
    if(model.videoCount < 1){
      // 延时初加载
      Timer(Duration(milliseconds: 3000), () {
        model.onPullToRefresh();
      });

    }else{
      // onPageSelected 会自动判断视频列表过小的话，loadmore
      model.onPageSelected(0);
      model.doLoadMore();
    }

  }

  @override
  void dispose() {
    ILog.lg.invoke();
    WidgetsBinding.instance.removeObserver(this);

    if (subscriptionNetwork != null) {
      subscriptionNetwork.cancel();
      subscriptionNetwork = null;
    }
    if (_timerUserGuid != null) {
      _timerUserGuid.cancel();
      _timerUserGuid = null;
    }
    super.dispose();
  }

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    ILog.lg.i("bbb- didChangeAppLifecycleState $state");
    if (state == AppLifecycleState.inactive) {
      isHiden = true;
      MainModel.of(context).onPause();
    } else if (state == AppLifecycleState.resumed) {
      isHiden = false;
      MainModel.of(context).onResume();
    }
  }

  @override
  Widget build(BuildContext context) {
    ILog.lg.invoke();
    if (MediaQuery.of(context).padding.bottom > 1) {
      Constant.FRINGESCREEEN = true;
    } else {
      Constant.FRINGESCREEEN = false;
    }
    return WillPopScope(
      child: Scaffold(
        resizeToAvoidBottomInset: false,
        resizeToAvoidBottomPadding: false,
//        body:
//        ScopedModelDescendant<MainModel>(builder: (context, child, model) {
//          return RefreshIndicator(
//              onRefresh: () async {
//                await MainModel.of(context).onPullToRefresh();
//                return Future.value(true);
//              },
//              child: _buildRefreshChild(model));
//        }),
        body:
        ScopedModelDescendant<MainModel>(builder: (context, child, model) {
          return RefreshIndicator(
              onRefresh: () async {
                await MainModel.of(context).onPullToRefresh();
                return Future.value(true);
              },
              child: _buildRefreshChild(model));
        }),
      ),
      onWillPop: () async {
        ILog.lg.i("onWillPop invoked ...");
        MainModel.of(context).dispose();
        if (subscriptionNetwork != null) {
          subscriptionNetwork.cancel();
          subscriptionNetwork = null;
        }
        return Future.value(true);
      },
    );
  }

  Widget _buildRefreshChild(MainModel model) {
    Widget _wBg = Container(
      color: Colors.black,
    );
    Widget _wContent;
    if (model.isNoNet) {
      // 首次进入，无网络 ...
      _wContent = VideoPageError(VideoPageError.STATE_NONET, _onClickRetry);
    } else if (model.hasError) {
      // 加载出现错误
      _wContent = VideoPageError(VideoPageError.STATE_ERROR, _onClickRetry);
    } else if (model.isRefreshing) {
      // 首次进入，加载中 ...
      _wContent = VideoPageError(VideoPageError.STATE_LOADING, _onClickRetry);
    } else {
      _wContent = _buildVideoContent(context, model);
    }

    return Stack(
      alignment: AlignmentDirectional.topStart,
      children: [_wBg, _wContent],
    );
  }

  void _onNetworkStates(ConnectivityResult result) {
    ILog.lg.i("_onNetworkStates: $result");
    if (result == ConnectivityResult.mobile && _networkToastShown == false) {
      _networkToastShown = true;
      SgToast.show("正在使用移动网络播放");
    }
  }

  Widget _buildVideoContent(BuildContext context, MainModel model) {
    Widget pageWidget = SgPageView.builder(
      controller: _pageController,
      itemBuilder: (context, position) => _buildPlayerItem(position, model),
      itemCount: (model.videoCount + 1), // 末尾增加一个 loading 页，请求更多
      onPageChanged: (int position) {
        MainModel mm = MainModel.of(context);
        mm.onPageSelected(position);
      },
      scrollDirection: Axis.vertical,
    );
    // 用户引导
    if (model.needUserGuid()) {
      if (Platform.isIOS) {
        Map map = Map();
        map["education"] = true;
        SgOperation.educationPageShow(map).then((value) {
          if (value != null && value == 1) {
            //说明ios层左滑手势启动了需要关闭教育引导
            model.resetUserGuid(false);
            _timerUserGuid.cancel();
            _timerUserGuid = null;
          }
        });
      }
      return GestureDetector(
        child: Stack(
          alignment: AlignmentDirectional.center,
          children: <Widget>[
            pageWidget,
            Container(
              child: UserGuidWidget(),
              color: Colors.transparent,
            )
          ],
        ),
        onTapDown: (TapDownDetails details) {
          model.resetUserGuid(false);
        },
        onPanDown: (details) {
          model.resetUserGuid(false);
        },
      );
    } else {
      return pageWidget;
    }
  }

  // 首次进入/下拉刷新失败后滴啊用
  void _onClickRetry() {
    MainModel.of(context).onClickRefresh();
  }

  // loadmore 失败后调用
  void _onClickReloadMore() {
    MainModel.of(context).onClickReloadMore();
  }

  Widget _buildPlayerItem(int position, MainModel model) {
    ILog.lg.invoke("Scrollable::_buildPlayerItem");
    if(position >= model.videoCount){
      Widget _wContent ;
      // 最后一个应该是加载中页面
      if (model.loadMoreNoNet) {
        _wContent = VideoPageError(VideoPageError.STATE_NONET, _onClickReloadMore);
      } else if (model.loadMoreError) {
        // 加载出现错误
        _wContent = VideoPageError(VideoPageError.STATE_ERROR, _onClickReloadMore);
      } else {
        // 加载中 ...
        _wContent = VideoPageError(VideoPageError.STATE_LOADING, _onClickReloadMore);
      }
      return _wContent;
    }else{
      var scopedMode = ScopedModel<VideoItemModel>(
        model: model.getVideoItemModel(position),
        child: VideoPageItem(),
      );
      return scopedMode;
    }
  }
}
