import 'dart:convert';
import 'dart:io';

import 'package:connectivity/connectivity.dart';
import 'package:flutter/material.dart';
import 'package:scoped_model/scoped_model.dart';
import 'package:short_video/base/i_log.dart';
import 'package:short_video/base/i_store_sp.dart';
import 'package:short_video/bean/video_item.dart';
import 'package:short_video/model_item.dart';
import 'package:short_video/network/sg_request.dart';
import 'package:short_video/platform/sg_operation.dart';
import 'package:short_video/player_pool.dart';
import 'package:short_video/util/statistics_helper.dart';
import 'package:short_video/util/toast.dart';
import 'package:short_video/widget/common/sg_page_view.dart';

import 'bean/rsp_video_list.dart';

/// 处理整体逻辑
class MainModel extends Model {
  // 后端配置，自动播放下一曲
  static final int ORDER_TYPE_AUTO = 2;

  // 后端配置，需要用户手动切换下一曲
  static final int ORDER_TYPE_MANUAL = 1;

  // 本地有效，忽略后端配置：写评论时，不再切换视频，但是需要循环播放当前视频
  static final int ORDER_TYPE_LOOP_FOR_COMMENTS = 3;

  // 控制播放顺序
  int _playOrder;

  static String kIsFirstUseShortVideo =
      "kIsFirstUseShortVideo"; //控制是否显示用户交易widget UI
  bool _isRefreshing = false;
  bool _hasError = false;
  bool _isNoNet = false;
  bool _needUserGuid;

  // 用来指示 load more
  bool _isLoadingMore = false;
  bool _loadMoreError = false;
  bool _loadMoreNoNet = false;

  int _currVideoIndex = 0;

  List<VideoItemModel> _videoModelList;

  SgPageController _pageController;

  bool get isRefreshing => _isRefreshing;

  /// 是否发生网络错误等
  bool get hasError => _hasError;

  /// 是否有网络
  bool get isNoNet => _isNoNet;

  /// 是否有网络
  bool get loadMoreError => _loadMoreError;
  /// 是否有网络
  bool get loadMoreNoNet => _loadMoreNoNet;

  void resetPageController(SgPageController spc) {
    _pageController = spc;
  }

  bool needUserGuid() {
    if (_needUserGuid == null) {
//      _needUserGuid = IStoreSp.sp.getBool(kIsFirstUseShortVideo);
      if (_needUserGuid == null) {
        _needUserGuid = true;
      }
    }
    return _needUserGuid;
  }

  void resetUserGuid(bool show) {
    assert(show != null);
    _needUserGuid = show;
//    IStoreSp.sp.setBool(kIsFirstUseShortVideo, _needUserGuid);
    notifyListeners();
  }

  int get videoCount => _videoModelList.length;

  VideoItemModel getVideoItemModel(int position) {
    if (_videoModelList.length <= position) {
      return null;
    } else {
      return _videoModelList[position];
    }
  }

  void setPlayOrder(int order) {
    _playOrder = order;
    getVideoItemModel(_currVideoIndex).isLooping =
        order == ORDER_TYPE_LOOP_FOR_COMMENTS || order == ORDER_TYPE_MANUAL;
  }

  int getPlayOrder() {
    return _playOrder;
  }

  MainModel(String json, this._playOrder) {
    _videoModelList = List<VideoItemModel>();
    _isRefreshing = true;
    if (json != null && json.length > 0) {
      Map<String, dynamic> videoMap = jsonDecode(json);
      if (videoMap != null) {
        VideoItem videoItem = VideoItem();
        videoItem.initFromJson(videoMap);
        if (videoItem.video_url != null && videoItem.video_url.length > 0) {
          _videoModelList.add(VideoItemModel(videoItem, 0, _onPlayComplete));
          _isRefreshing = false;
        }
      }
    }
    ILog.lg.i("_isRefreshing:$_isRefreshing");
  }

  /// 下拉刷新
  void onPullToRefresh() async {
    List<VideoItem> videos;
    ReqGetVideoList req = ReqGetVideoList();
    RspVideoList rsp =  await req.doRequest();
    if(rsp.isServerOK()){
      videos = rsp.videos;
      _updateItemModelList(videos, false);
    }else{
      // 网络出错
      // 判断是否无网络
      _hasError = true;
      ConnectivityResult result = await Connectivity().checkConnectivity();
      if (result == ConnectivityResult.none) {
        _isNoNet = true;
      } else {
        _isNoNet = false;
      }
      // 刷新失败，需要把旧的列表清除，否则会有正在播放的声音；
      _updateItemModelList(null, false);
    }

    _isRefreshing = false;

    notifyListeners();
    if (!_hasError) {
      onPageSelected(0);
    }
  }

  /// 网络出错，点击刷新
  void onClickRefresh() async {
    _isRefreshing = true;
    _isNoNet = false;
    _hasError = false;
    // 更新刷新界面
    notifyListeners();
    ConnectivityResult result = await Connectivity().checkConnectivity();
    if (result == ConnectivityResult.none) {
      _isNoNet = true;
      _isRefreshing = false;
    } else {
      _isNoNet = false;

      List<VideoItem> videos;
      ReqGetVideoList req = ReqGetVideoList();
      RspVideoList rsp =  await req.doRequest();
      if(rsp.isServerOK()){
        videos = rsp.videos;
        _updateItemModelList(videos, false);
      }else{
        // 网络出错
        _hasError = true;
        ConnectivityResult result = await Connectivity().checkConnectivity();
        if (result == ConnectivityResult.none) {
          _isNoNet = true;
        } else {
          _isNoNet = false;
        }
      }
      _isRefreshing = false;
    }
    notifyListeners();
    if (!_isNoNet && !_hasError) {
      onPageSelected(0);
    }
  }

  void onClickReloadMore() async {
    _loadMoreError = false;
    _loadMoreNoNet = false;
    _isLoadingMore = false;

    notifyListeners();

    await doLoadMore();
  }

  // 加载到底部，执行加载更多
  Future<void> doLoadMore() async {

    List<VideoItem> videos;
    ReqGetVideoList req = ReqGetVideoList();
    RspVideoList rsp =  await req.doRequest();
    if(rsp.isServerOK()){
      videos = rsp.videos;
      _updateItemModelList(videos, true);
    }else{
      _loadMoreError = true;
      // 网络出错
      ConnectivityResult result = await Connectivity().checkConnectivity();
      if (result == ConnectivityResult.none) {
        _loadMoreNoNet = true;
      } else {
        _loadMoreNoNet = false;
      }
    }

    ILog.lg.i("doLoadMore::notifyListeners::onPageSelected:$_currVideoIndex  _pageController.page: ${_pageController.page}");

    int targetPage = _pageController.page.floor();
    if(targetPage != _currVideoIndex || _currVideoIndex == 0){
      // 首页进来也要预取
      onPageSelected(targetPage);
    }

    notifyListeners();
  }

  /// start from '0'
  void onPageSelected(int pageIndex) {
    ILog.lg.i("onPageSelected:$pageIndex, ${_videoModelList.length}");
    if (_videoModelList == null) {
      return;
    }
    if (pageIndex >= _videoModelList.length) {
      // 滑到最后一个，加载中。。。，清理其它页面状态
      for(int i = 0 ; i < _videoModelList.length; i  ++){
        _videoModelList[i].onPageChanged(_currVideoIndex+1);
      }
      return;
    }
    // 执行load more
    if (pageIndex == (_videoModelList.length - 2) && !_isRefreshing) {
      // 滚动到倒数第2条时，加载更多
      doLoadMore();
    }

    //当视频切换时通知native
    if(Platform.isAndroid){
      SgOperation.switchVideo(Map());
    }

    _currVideoIndex = pageIndex;

    if (ORDER_TYPE_LOOP_FOR_COMMENTS == _playOrder ||
        ORDER_TYPE_MANUAL == _playOrder) {
      _videoModelList[_currVideoIndex].isLooping = true;
    } else {
      _videoModelList[_currVideoIndex].isLooping = false;
    }

    for(int i = 0 ; i < _videoModelList.length; i  ++){
      _videoModelList[i].onPageChanged(_currVideoIndex);
    }
  }

  void _onPlayComplete(int index) {
    ILog.lg.i("_onPlayComplete:_playOrder$_playOrder, index:$index, _pageController:$_pageController _currVideoIndex:$_currVideoIndex _videoModelList:${_videoModelList.length}");

    if (ORDER_TYPE_AUTO == _playOrder) {
      // 自动播放下一曲
      if (_pageController != null && _currVideoIndex < _videoModelList.length) {
        _pageController
            .nextPage(duration: Duration(milliseconds: 500), curve: Curves.ease);
      }else{
        // 没有最后一条了
        _videoModelList[_currVideoIndex].applyPause();
      }
    }
  }

  void onPause() async {
    ILog.lg.i("model_main:onPause");
    // 循环遍历所有的 item，不可只遍历当前；因为有可能当前被切到下一条，造成状态不一致
    if (_videoModelList != null) {
      _videoModelList.forEach((item){
        item.onPause();
      });
    }
  }

  void onResume() async {
    ILog.lg.i("model_main:onResume");
    if (_videoModelList != null) {
      _videoModelList.forEach((item){
        item.onResume();
      });
    }
  }

  void dispose() async {
    ILog.lg.i("model_main:dispose");
    if (_videoModelList != null) {
      _videoModelList.forEach((itemModel) {
        itemModel.dispose();
      });
      _videoModelList.clear();
    }
    PlayerPool.clear();
  }

  /// doAppend: 标示是否'追加'，否则替换当前列表
  /// 注意：记得清空列表时，通知itemmode 销毁事件
  void _updateItemModelList(List<VideoItem> videos, bool doAppend) {
    if (!doAppend) {
      // 清理上次记录
      _videoModelList.forEach((itemModel) {
        itemModel.dispose();
      });
      _videoModelList.clear();
    }
    // 将本次记录追加到列表中
    if (videos != null && videos.length > 0) {
      int oldListLen = _videoModelList.length;
      List<String> urls = List();
      for (int i = 0; i < videos.length; i++) {
        _videoModelList
            .add(VideoItemModel(videos[i], i + oldListLen, _onPlayComplete));
            urls.add(videos[i].video_url);
      }
      SgOperation.cacheVideo(urls, true);
    }
  }

  static MainModel of(BuildContext ctx) {
    return ScopedModel.of<MainModel>(ctx);
  }
}
