import 'dart:async';
import 'dart:convert';
import 'dart:math';

import 'package:short_video/base/i_log.dart';
import 'package:short_video/base/i_net_http.dart';
import 'package:short_video/build_config.dart';
import 'package:short_video/network/i_rsp_base.dart';
import 'package:short_video/network/net_const.dart';
import 'package:short_video/network/root_url.dart';

/// 封装请求发送操作；
/// 处理token失效以及部分公共code
abstract class ISgRequest<T extends BaseResponse> {
  /// 发送请求
  Future<T> doRequest() async {
    var targetURL = host() + apiPath();
    var para = pkgParameter();
    Completer<T> cmpRes = Completer<T>();
    INetHttp_Response httpResponse;

    T resObj = createResponseObj();
    try{
      httpResponse= await makeRemoteRequest(targetURL, para);
    }catch(e){
      // 有可能请求出错
      resObj.code = NetConst.HTTP_ERROR_SERVER_UNREACHABLE;
      resObj.msg = "Can not make connection to server: $e";
      ILog.lg.e(resObj.msg);
      cmpRes.complete(resObj);
      return cmpRes.future;
    }

    if (BuildConfig().isInDebugMode()) {
      ILog.lg.i("$targetURL -> \r\n ${httpResponse.statusCode}");
    }

    Map<String, dynamic> jsonResponse;
    int errorCode;
    String errorMsg;
    if (httpResponse != null && httpResponse.data != null) {
      // 默认只支持utf8 编码
      String jsonStr;
      try {
        jsonStr = utf8.decode(httpResponse.data);
      } catch (e) {
        errorCode = NetConst.HTTP_ERROR_CODEC_WRONG;
        errorMsg = "Can not decode string from data by utf-8:: ${httpResponse.headers['contentType']}";
      }
      try {
        jsonResponse = json.decode(jsonStr);
      } catch (e) {
        errorCode = NetConst.HTTP_ERROR_WRONG_FORMATION;
        errorMsg = "Parse reponse object error $e";
      }
    } else {
      errorCode = NetConst.HTTP_ERROR_WRONG_CONTENT_TYPE;
      errorMsg = "Unknow ContentType: ${httpResponse.headers['contentType']} ; only support for json/text";
    }
    if(errorCode == null && jsonResponse != null){
      resObj.initFromJson(jsonResponse);
    }else{
      resObj.code = errorCode;
      resObj.msg = errorMsg;
    }
    cmpRes.complete(resObj);
    return cmpRes.future;
  }

  Future<INetHttp_Response> makeRemoteRequest(
      String targetURL, Map<String, dynamic> para) {
    assert(INetHttp.http != null);
    return INetHttp.http
        .makeHttpRequest(targetURL, para, null, isGet());
  }

  Future<INetHttp_Response> doFormRequest() {
    ILog.lg.invoke();

    var targetURL = this.host();
    targetURL = targetURL + this.apiPath();
    var para = pkgParameter();

    ILog.lg.i("targetURL:$targetURL");
    assert(INetHttp.http != null);
    return INetHttp.http.makeFormRequest(targetURL, para, null);
  }

  /// api host
  String host() {
    ILog.lg.i("host ...");
    return "https://www.baidu.com/";
  }

  /// api path
  String apiPath() {
    return "apitemp/";
  }

  bool isGet(){
    return true;
  }

  /// 防止循环获取token
  int tokenRetry = 3;

  /// 所需参数
  Map<String, dynamic> pkgParameter() {
    var para = Map<String, dynamic>();
    para['ts'] = DateTime.now().millisecondsSinceEpoch.toString();
    return para;
  }

  /// 直接返回新的 ISgResonse 对象即可
  T createResponseObj();
}

/// 业务API
abstract class ISgReqBusiness extends ISgRequest {
  static Map<dynamic, dynamic> paraCommon;
  @override
  String host() {
    ILog.lg.i("ISgReqBusiness host ...");
    return UrlRoot.root(RootType.BUSINESS);
  }

  @override
  Map<String, dynamic> pkgParameter() {
    Map<String, dynamic> superMap = super.pkgParameter();
    assert(superMap != null);
    if(paraCommon != null){
      paraCommon.forEach((key, value){
        superMap[key] = value;
      });
    }
    return addBussParameter(superMap);
  }

  /// 实际的需要请求的业务逻辑相关的参数，如无需另外的参数，可直接返回 commonParameter 即可
  Map<String, dynamic> addBussParameter(Map<String, dynamic> commonParameter) {
    return commonParameter;
  }
}

class SgErrorHttp extends Error {
  final String message;
  final int code;

  SgErrorHttp(this.code, this.message);

  String toString() => "$code -> $message";
}
