import 'dart:async';
import 'dart:convert';
import 'dart:typed_data';

import 'package:dio/dio.dart';
import 'package:short_video/base/i_log.dart';

import 'dart:io';

import 'package:short_video/base/i_net_http.dart';
import 'package:short_video/build_config.dart';
import 'package:short_video/network/root_url.dart';

/// 单例
class SogouClient extends INetHttp {
  static SogouClient _singleton = SogouClient._internal();
  static Dio _dio;

  SogouClient._internal() {
    ILog.lg.i("SogouClient created");
    _initDio();
  }

  factory SogouClient() {
    return _singleton;
  }

  bool _initDio() {
    _dio = new Dio();
    _dio.options.connectTimeout = 5000; //5s
    _dio.options.receiveTimeout = 5000;
    DefaultHttpClientAdapter clientAdapter = DefaultHttpClientAdapter();
    clientAdapter.onHttpClientCreate = (HttpClient client) {
      ILog.lg.i("dio.httpClientAdapter.onHttpClientCreate invoked");
      // config the http client
      if (BuildConfig().needDebugMenu()) {
        if (UrlRoot.proxyEnable()) {
          var proxyHost = UrlRoot.getProxyHost();
          var proxyPort = UrlRoot.getProxyPort();
          client.findProxy = (uri) {
            //proxy all request to localhost:8888
            var proxyString = "PROXY $proxyHost:$proxyPort";
            ILog.lg.d("PROXY:" + proxyString);
            return proxyString;
          };
        }
        // 处理 charles 抓包时无法建立成功的问题
        client.badCertificateCallback =
            (X509Certificate cert, String host, int port) => true;
      }
      return client;
    };
    _dio.httpClientAdapter = clientAdapter;
    return true;
  }

  Future<Response> uploadFile(
      String targetURL,
      String localFilePath,
      Uint8List localData,
      String remoteFileName,
      Map<String, String> commonPara) {
    assert(_dio != null);
    assert((localFilePath != null && localFilePath.length > 0) ||
        localData != null);
    assert(remoteFileName != null && remoteFileName.length > 0);
    assert(targetURL != null && targetURL != null);
    ILog.lg.i("NEW REQ:" + targetURL);

    print('PicLocalPath:' + localFilePath);

    FormData formData;

    if (localData != null) {
      formData = new FormData.from({
        "file": new UploadFileInfo.fromBytes(localData, remoteFileName),
      });
    } else {
      formData = new FormData.from({
        "file": new UploadFileInfo(File(localFilePath), remoteFileName,
            contentType: ContentType("image", "jpeg", charset: "utf-8")),
      });
    }

    commonPara.forEach((eleKey, eleValue) {
      formData.add(eleKey, eleValue);
    });

    ILog.lg.i('upload : ' + formData.toString());

    try {
      return _dio.post(targetURL, data: formData);
    } catch (e) {
      ILog.lg.i(e.toString());
      return null;
    }
  }

  @override
  Future<INetHttp_Response> makeHttpRequest(
      String targetURL, Map<String, dynamic> para, Map<String, dynamic> headers,
      [bool isGet]) async {
    assert(_dio != null);
    assert(targetURL != null && targetURL != "");
    ILog.lg.i("NEW REQ:" + targetURL);

    _pkgHttpHeader(headers);
    Response dioRsp;
    // send request to server
    try {
      // Create a FormData
      if (para != null) {
        if (isGet != null && isGet) {
          dioRsp = await _dio.get(targetURL, queryParameters: para);
        } else {
          var jsonStr = json.encode(para);
          dioRsp = await _dio.post(targetURL,
              data: jsonStr,
              options: new Options(
                  contentType:
                      ContentType.parse("application/json;charset=UTF-8")));
        }
      } else {
        if (isGet != null && isGet) {
          dioRsp = await _dio.get(targetURL);
        }else{
          dioRsp = await _dio.post(targetURL);
        }
      }
    } catch (e) {
      ILog.lg.e(e.toString());
      INetHttp_Response errorRsp = INetHttp_Response();
      errorRsp.statusCode = -100;
      errorRsp.data = utf8.encode(e.toString());
      return errorRsp;
    }

    return _pkgRsp(dioRsp);
  }

  void onHttpPostError(e) {}

  /// 返回类型只接收 二进制流
  @override
  Future<INetHttp_Response> makeFormRequest(String targetURL,
      Map<String, dynamic> para, Map<String, dynamic> headers) async {
    assert(_dio != null);
    assert(targetURL != null && targetURL != "");
    assert(para != null && para.length > 0);
    // 需要处理转换一下，将 Uint8List -> UploadFileInfo
    Map<String, dynamic> targetPara = Map<String, dynamic>();

    if (para != null) {
      para.forEach((eleKey, eleValue) {
        if (eleValue is Uint8List) {
          var uploadInfo = UploadFileInfo.fromBytes(eleValue, "value_name",
              contentType: ContentType.binary);
          targetPara[eleKey] = uploadInfo;
        } else {
          targetPara[eleKey] = eleValue;
        }
      });
    }

    ILog.lg.i("NEW REQ:" + targetURL);

    _pkgHttpHeader(headers);

    FormData formData = FormData.from(targetPara);

    Response dioRsp = await _dio.post(
      targetURL,
      data: formData,
    );

    return _pkgRsp(dioRsp);
  }

  void _pkgHttpHeader(Map<String, dynamic> headers) {
    if (headers != null) {
      headers.forEach((eleKey, eleValue) {
        _dio.options.headers[eleKey] = eleValue;
      });
    }
  }

  INetHttp_Response _pkgRsp(Response dioRsp) {
    INetHttp_Response res = INetHttp_Response();
    res.statusCode = dioRsp.statusCode;
    res.headers = Map<String, String>();
    dioRsp.headers.forEach((String name, List<String> values) {
      if (values != null && values.length > 0) {
        String vStr;
        values.forEach((vItem) {
          if (vStr != null) {
            vStr += ";";
            vStr += vItem;
          } else {
            vStr = vItem;
          }
        });
        res.headers[name] = vStr;
      }
    });
    if (dioRsp.data is String) {
      res.data = utf8.encode(dioRsp.data);
    } else if (dioRsp.data is Map<String, dynamic>) {
      res.data = utf8.encode(json.encode(dioRsp.data));
    } else if (dioRsp.data is List) {
      res.data = dioRsp.data;
    } else {
      ILog.lg.e("Unknown data type:${dioRsp.data}");
    }

    return res;
  }
}
